#include <math.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
//#include <malloc.h>

//#include "hmm.h"
//#include "ss_maps.h"
//#include "hmm_model_tools.h"
//#include "eval_struct.h"
//#include "conversion.h"

#define MNratio 1.00
#define Eratio 1.40
#define Lratio 1.00


// qlambda_ramaland is a version of qlambda_ramacen that
//                  chooses hierachially which centroid
//                  to predict. First votes are tallied
//                  for 'continents' H E L X, then the winning
//                  continent chooses a 'country'.
//                  For better prediction of angles in the
//                  border regions, there are overlapping
//                  continents HE and EX.
// CB 5-DEC-98

void qlambda_ramaland ( 
	      struct parent pa, 
	      struct model mod , 
	      int *q_ptr, 
	      float *cf, 
	      double **gamma)
{
  int i, j, t, k, TT, NN ;
  int iq, iv ;
  float P;
  double vv, vote3[MRAMA];
  int voterama_continent(int , float** , double *, double *)  ;
  int voterama_region1(int , float** , double *, double *) ;
  int voterama_region0(int , float** , double *, double *) ;



  TT = pa.T ; 
  NN = mod.N  ; 


  // VOTE!
  for ( t = 0 ; t < TT ; t++ ){
    // vote for a continent
    iv = voterama_continent(NN,mod.r, gamma[t], &vv) ;

    if (iv == 0 )     {  // Ll
      cf[t] = vv ;   // confidence of {Ll}
      iv = voterama_region0(NN,mod.r, gamma[t], &vv); 
    }
    else if (iv == 1) { // all phi < 0 regions
      iv = voterama_region1(NN,mod.r, gamma[t], &vv); 
      cf[t] = vv ;  // confidence of subregion 
    } 
    else if (iv == 2) { 
      cf[t] = vv ;   // confidence of {c}
      iv = MRAMA-1;  // last ramacen is c=cis 
    } 
     
    q_ptr[t] = iv ;
  }

}

int voterama_continent(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

 // continent 0 Ll       =  r[7] + r[8]
 // continent 1 HGEBdbex =  r[0] + r[1] + r[3] + r[4] + r[5] + r[6] + r[9]
 // continent 2 c        =  r[10]

  for (i=0; i<3; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    x = r[i][7] + r[i][8] ;
    vote[0] += x * gamma[i] ;
    x = r[i][0] + r[i][1] + r[i][2] + r[i][3] + r[i][4] + 
        r[i][5] + r[i][6] + r[i][9] ;
    vote[1] += x * gamma[i] ;
    x = r[i][10] ;
    vote[2] += x * gamma[i] ;
  }
  // get most popular region 'iv'
  vv = 0. ; 
  for (j = 0; j < 3 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;
  return iv ;
}

// rama region0 = Ll
// returns 7,8 or -1

int voterama_region0(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

 // country 0 L       =  r[7] 
 // country 1 l       =  r[8]

  for (i=0; i<2; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    vote[0] += r[i][7] * gamma[i] ;
    vote[1] += r[i][8] * gamma[i] ;
  }
  vv = 0. ; 
  for (j = 0; j < 2 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;

  if (iv == 0) return 7;
  if (iv == 1) return 8;

  return -1 ;
}

// region1 = HGEBdbex

int voterama_region1(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;
  int voterama_GH(int , float **, double *, double *);
  int voterama_Gdb(int , float **, double *, double *);
  int voterama_EBdb(int , float **, double *, double *);
  int voterama_xeE(int , float **, double *, double *);

 // region 0 HG      =  r[0] + r[1]
 // region 1 Gdb     =  r[1] + r[4] + r[5]
 // region 2 dbEB    =  r[2] + r[3] + r[4] + r[5]
 // region 3 xeE     =  r[3] + r[6] + r[9]

  for (i=0; i<4; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    x = r[i][0] + r[i][1] ;
    vote[0] += x * gamma[i] ;
    x = r[i][1] + r[i][4] + r[i][5];
    vote[1] += x * gamma[i] ;
    x = r[i][2] + r[i][3] + r[i][4] + r[i][5] ;
    vote[2] += x * gamma[i] ;
    x = r[i][3] + r[i][6] + r[i][9] ;
    vote[3] += x * gamma[i] ;
  }
  // get most popular region 'iv'
  vv = 0. ; 
  for (j = 0; j < 4 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  if (iv < 0) return iv ;
  *cf = vv ;   // confidence of region
  if (iv == 0)      { iv = voterama_GH(NN, r, gamma, &vv) ; }
  else if (iv == 1) { iv = voterama_Gdb(NN, r, gamma, &vv) ; }
  else if (iv == 2) { iv = voterama_EBdb(NN, r, gamma, &vv) ; }
  else if (iv == 3) { iv = voterama_xeE(NN, r, gamma, &vv) ; }
  // *cf = vv ; confidence of centroid (not used)
  return iv ;
}

int voterama_Gdb(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

  for (i=0; i<3; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    vote[0] += r[i][1] * gamma[i] ;
    vote[1] += r[i][4] * gamma[i] ;
    vote[2] += r[i][5] * gamma[i] ;
  }
  vv = 0. ; 
  for (j = 0; j < 3 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;
  if (iv == 0) return 1;
  if (iv == 1) return 4;
  if (iv == 2) return 5;
  return -1 ;
}

int voterama_xeE(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

  for (i=0; i<3; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    vote[0] += r[i][3] * gamma[i] ;
    vote[1] += r[i][6] * gamma[i] ;
    vote[2] += r[i][9] * gamma[i] ;
  }
  vv = 0. ; 
  for (j = 0; j < 3 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;
  if (iv == 0) return 3;
  if (iv == 1) return 6;
  if (iv == 2) return 9;
  return -1 ;
}

int voterama_EBdb(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

  for (i=0; i<4; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    vote[0] += r[i][2] * gamma[i] ;
    vote[1] += r[i][3] * gamma[i] ;
    vote[2] += r[i][4] * gamma[i] ;
    vote[3] += r[i][5] * gamma[i] ;
  }
  vv = 0. ; 
  for (j = 0; j < 4 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;
  if (iv == 0) return 2;
  if (iv == 1) return 3;
  if (iv == 2) return 4;
  if (iv == 3) return 5;
  return -1 ;
}

int voterama_GH(int NN, float** r, double *gamma, double *cf) 
{
  double vv, vote[MRAMA];
  float x;
  int i,j,iv=-1;

  for (i=0; i<2; i++) vote[i] = 0.;
  for ( i = 1 ; i < NN ; i++ ){
    vote[0] += r[i][0] * gamma[i] ;
    vote[1] += r[i][1] * gamma[i] ;
  }
  vv = 0. ; 
  for (j = 0; j < 2 ; j++) {
    if (vote[j] > vv) {vv = vote[j]; iv = j ;} 
  }
  *cf = vv ;
  if (iv == 0) return 0;
  if (iv == 1) return 1;
  return -1 ;
}


//
//					September 16, 1997
//
//    max.C - given a[N], find max(a[i]) and i at max 
//

void max ( float a[], int N,  float *max, int *imax)
{
  int i;
   
  *max = a[0];
  *imax = 0;
 
  for ( i = 1 ; i < N ; i++){
    if ( a[i] >= *max ){
      *max = a[i];
      *imax = i;
    }
  }

}


//
//					September 16, 1997
//
//    dmax.C - given a[N], find max(a[i]) and i at max 
//

void dmax ( double a[], int N,  double *max, int *imax)
{
  int i;
   
  *max = a[0];
  *imax = 0;
 
  for ( i = 1 ; i < N ; i++){
    if ( a[i] >= *max ){
      *max = (float) a[i];
      *imax = i;
    }
  }

}
