//*=========================================================================
//** Copyright 2003 Chris Bystroff
//**
//** Permission is hereby granted to use, execute, copy, distribute, modify,
//** and distribute in modified form this software for non-profit purposes,
//** provided this notice is retained in any and all copies.
//**
//** For for-profit usage, please contact bystrc@rpi.edu
//*=========================================================================
// getalphabeta_F.C
// This module computes the forward and backward algorithms
// This version receives a vector of flags, each flag determines
// which data are used at a given position in the sequence.
//  8-dec-99  CB
#include <fstream.h>
#include <stdlib.h>
#include <iostream.h>
//#include "hmm.h"
//#include "conversion.h"
//#include "ss_maps.h"
//#include "hmm_model_tools.h"
//#include "eval_struct.h"

int O[Tmax];
double brat, rrat, drat, ss;

//
//   				         Feb 18, 1999
//
//     getalpha_brdc_prod.C - Compute alpha, using b, r, c, d vectors  
//                       uses FLAG to deteremine which emissions to use
//                       Use bprof() in scoring b.
//
//                       Uses terminal and initial distributions 
//                                               
void getalpha_brdc_prod ( 
	       double **alpha, 
	       struct parent pa, 
	       struct model md, 
	       int **itr ,
	       double *ct ,
               int *iflag )
{
  int N, T, i, j, t, ind ;
  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O );

  // t = 0 

  drat = 1.0; rrat = 1.0 ;

  // ========================== alpha for states @ 0
  ct[0] = 0.;
  for ( j = 1 ; j < N ; j++ ){ 
    alpha[0][j] = md.a[0][j] * flagrat_brdc_prod ( iflag[0], j, 0, O, pa, md ) ;
    ct[0] += alpha[0][j];
  }
  // cout << "position  0 in alpha: " << ct[0] << endl;
  ct[0] = 1./ct[0] ; // sum of alpha[0]
  for ( j = 1; j < N; j++ ) alpha[0][j] *= ct[0] ;
  // ========================== alpha for NOT-state @ 0
  j = 0;  
  ss = 0.;
  for ( i = 0; i < itr[j][0] ; i++ ){   
    ind = itr[j][i+1] ;
    ss += alpha[0][ind] * md.a[ind][j] ;
  }
  alpha[0][j] = ss ;
  for ( t = 1; t < T; t++ ){
    // ========================== alpha for states @ t
    ct[t] = 0.;  // init scale factor
    for ( j = 1; j < N; j++ ){
      ss = 0.;
      for ( i = 0; i < itr[j][0] ; i++ ){   
        ind = itr[j][i+1] ;
	ss += alpha[t-1][ind] * md.a[ind][j] ;
      }
      alpha[t][j] = ss *  flagrat_brdc_prod ( iflag[t], j, t, O, pa, md ) ;
      ct[t] += alpha[t][j] ;  // inverse of scale factor for next cycle
    }
    // ========================== error if sum(alpha) == 0.
    if ( ct[t] == 0 ){
      cout << "Error: no incoming transition found for position " << t+1 <<
	" in family " << pa.code << pa.chainID << pa.domainID << endl ; 
      cout << "Position " << t << " Residue: " << pa.O_ptr[t-1] <<
	" DSSP symbol: " << pa.O_ss[t-1] << 
	" Rama symbol: " << pa.O_rama[t-1] <<
#ifdef USE_CONTEXT 
	" Context: " << pa.O_context[t-1] << 
#endif
	endl ;
      cout << "Position " << t+1 << " Residue: " << pa.O_ptr[t] <<
             " DSSP symbol: " << pa.O_ss[t] << 
	     " Rama symbol: " << pa.O_rama[t] <<
#ifdef USE_CONTEXT 
	     " Context: " << pa.O_context[t] << 
#endif
	     endl ;
      for (i = t; i < T; i++) ct[i] = 1.; // allow beta calc 
      return ;  // find other end of bad section

      //exit(8) ;
    } else {   // no error
      ct[t] = 1./ct[t] ;
      // cout << "alpha: position " <<   t << " : " << ct[t] << endl;
      for ( j = 1; j < N; j++ ) alpha[t][j] *= ct[t] ;
      // ========================== alpha for NOT-state @ t
      j = 0;  
      ss = 0.;
      for ( i = 0; i < itr[j][0] ; i++ ){   
        ind = itr[j][i+1] ;
        ss += alpha[t][ind] * md.a[ind][j] ;
      }
      alpha[t][j] = ss ;
    }

  }
}
//
//   				  Feb 18, 1999 
//
//     getbeta_brdc_prod.C - Compute beta using bprof
//

void getbeta_brdc_prod ( 
               double **beta, 
	       struct parent pa, 
	       struct model md,
	       int **otr ,
               double *ct,
               int *iflag  )
{
  int N, T, i, j, t, jnd ;
  float cct;

  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O ); 

  // ========================== beta for state i @ T
  // for ( i = 1 ; i < N ; i++ ) beta[T-1][i] = md.a[i][0] * ct[T-1] ; 
  for ( i = 1 ; i < N ; i++ ) beta[T-1][i] = ct[T-1] ; 
  // ========================== beta for NOT-state @ T
  // i = 0;
  ss = 0.;
  for ( j = 0; j < otr[0][0] ; j++ ) {
    jnd = otr[0][j+1] ;
    ss += md.a[0][jnd] * beta[T-1][jnd] *
	  flagrat_brdc_prod(iflag[T-1],jnd,T-1,O,pa,md) ;
  }
  beta[T-1][0] = ss ;   // do not scale beta for NOT-state

  for ( t = T-2 ; t >= 0 ; t-- ){
    // cout << "beta: position " <<   t << " : " << ct[t] << endl;
    cct = 0.;
    // ========================== beta for state i @ t
    for ( i = 1; i < N; i++ ){
      ss = 0.;
      for ( j = 0; j < otr[i][0] ; j++ ) {
        jnd = otr[i][j+1] ;
	ss += md.a[i][jnd] * beta[t+1][jnd] * 
	  flagrat_brdc_prod(iflag[t],jnd,t+1,O,pa,md) ;
      }
      beta[t][i] = ss * ct[t] ;
      cct += beta[t][i] ;
    }
    // ========================== beta for NOT-state  @ t
    // i = 0;   
    ss = 0.;
    for ( j = 0; j < otr[0][0] ; j++ ) {
      jnd = otr[0][j+1] ;
      ss += md.a[0][jnd] * beta[t][jnd] *
	  flagrat_brdc_prod(iflag[t],jnd,t,O,pa,md) ;
    }
    beta[t][0] = ss ;  // do not scale beta for NOT-state
    if ( cct == 0 ){
      cout << "Error: no outgoing transition found for position " << t+1 <<
	      " in family " << pa.code << pa.chainID << pa.domainID << endl ; 
      cout << "Position " << t << " Residue: " << pa.O_ptr[t-1] <<
	      " DSSP symbol: " << pa.O_ss[t-1] << 
	      " Rama symbol: " << pa.O_rama[t-1] <<
#ifdef USE_CONTEXT 
	      " Context: " << pa.O_context[t-1] << 
#endif
	      endl ;
      cout << "Position " << t+1 << " Residue: " << pa.O_ptr[t] <<
	      " DSSP symbol: " << pa.O_ss[t] << 
	      " Rama symbol: " << pa.O_rama[t] <<
#ifdef USE_CONTEXT 
	      " Context: " << pa.O_context[t] << 
#endif
	      endl ;
      exit(8) ;
    }

  }
}













