// 
// Dec 8, 1999
// 
// gamma.C - Outputs gamma matrix as dbase[na].T lines of length mod.N
// gammaCA.C ... *plus* distance matrix of C-alphas, T values for each
// line.
// Must be compiled with new version of vall_tools.C 
// and hmm.h which includes 'ca'
// as part of the "parent" structure.
// gammaCB.C -- Uses C-betas instead of C-alphas for the DISTANCE lines
// (For glycines, C-alpha is used)
// 
// 
// hmmstr_gamma.C 
// change 20-FEB-2003: Yu Shao: add the secondary structure probabilities
// for each
// residue position (keyword: SS)
// change 03-MAR-2003: fix a bug: don't calculate the distance for
// residues with no coordinates.
// change 09-FEB-2004: Yu Shao : fixed the null value of the structure_name 
// (4-letter pdb code) in vall_toos2.C.
// *=========================================================================
// ** Copyright 2003 Chris Bystroff
// **
// ** Permission is hereby granted to use, execute, copy, distribute,
// modify,
// ** and distribute in modified form this software for non-profit
// purposes,
// ** provided this notice is retained in any and all copies.
// **
// ** For for-profit usage, please contact bystrc@rpi.edu
// *=========================================================================
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdio.h>

#include "hmm.h"
#include "hmm_model_tools.h"
#include "vall_tools2.h"
#include "conversion.h"
#include "ss_maps.h"
#include "eval_struct.h"

#include "getalphabeta_F.C"
#include "eval_struct0.C"
#include "conversion.C"
#include "hmm_model_tools.C"
#include "ss_maps.C"
#include "vall_tools2.C"

#define CONTACT 7.0
#define MDACUT 100.0
#define GPROF 1

float           EPS = 0.01;
float           NCOUNT = 2.0;
float           ramacen[MRAMA][2];

main(int argc, char *argv[])
{

    float          *cfr;
    int            *q_ptr_r;
    char            rama_string[MRAMA + 1] = "HGBEdbeLlxc";
    char           *ss_max_prob;
    short int       T;
    int             i,
                    j,
                    t,
                    na,
                    N,
                    NNa,
                    ii,
                    intSS,
                    intQ,
                    intStructureFlag,
                    verbose=0;
    double        **alpha,
                  **beta,
                  **gamma,
                  **ss_prob;
    double         *ct,
                    dSum;
    float           g,
                    x,
                    z,
                    y[3];
    int           **intrans,
                  **outtrans;
    int            *iflag;
    struct model    mod;
    struct parent  *dbase;        // array of all parent structures
    void            getalpha_brdc_prod(double **, struct parent,
                                       struct model, int **, double *, int *);
    void            getbeta_brdc_prod(double **, struct parent,
                                      struct model, int **, double *,
                                      int *);

    if (argc < 5) {
        cout << "hmmstr_gamma.C  v.02-FEB-2004" << endl;
        cout << "gammaCB3_cath   v.03-MAR-2003" << endl;
        cout << "gammaCB3_cath   v.26-FEB-2003" << endl;
        // cout << "**** Use structure if available!! ****" << endl;
        cout << "argv[1] = HMMSTR model file" << endl;
        cout << "argv[2] = vall/drct file" << endl;
        cout << "argv[3] = output file (gca)" << endl;
        cout << "argv[4] = 0 (don't use structure), 1 (use structure)" << endl;
        cout << "argv[5] = cath domain name (optional)" << endl;
        cout << "argv[6] = cath code (optional)" << endl;
        return (0);
    }

    intStructureFlag = atoi(argv[4]);

    // prob_flag = ctoi( argv[3] );
    // na = ctoi ( argv[4] ) - 1 ;
    na = 0;

    // Read in the model 

    cout << "Reading the HMM file: " << argv[1] << "\n" << flush;
    modelf(argv[1], &mod);
    N = mod.N;


    // Initialize 

    ramacen[0][0] = -61.91;
    ramacen[0][1] = -45.20;        // H
    ramacen[1][0] = -109.78;
    ramacen[1][1] = 20.88;        // G
    ramacen[2][0] = -70.58;
    ramacen[2][1] = 147.22;        // B
    ramacen[3][0] = -132.89;
    ramacen[3][1] = 142.43;        // E
    ramacen[4][0] = -135.03;
    ramacen[4][1] = 77.26;        // d
    ramacen[5][0] = -85.03;
    ramacen[5][1] = 72.26;        // b
    ramacen[6][0] = -165.00;
    ramacen[6][1] = 175.00;        // e
    ramacen[7][0] = 55.88;
    ramacen[7][1] = 38.62;        // L
    ramacen[8][0] = 85.82;
    ramacen[8][1] = -0.03;        // l
    ramacen[9][0] = 80.;
    ramacen[9][1] = -170.00;        // x
    ramacen[10][0] = -70.0;
    ramacen[10][1] = 150.00;        // c=cis!!

    // Read in the database

    cout << "Reading the DRCT file: " << argv[2] << "\n" << flush;
    // Bug: program core dumps when the drct file is empty.
    // Solution: check the number of residues read from drct file,
    // if 0, output the error message and exit.
    // -- by Yu Shao (2003)
    // dbasef1 ( argv[2], &dbase, &NNa, &iflag );
    if (dbasef1(argv[2], &dbase, &NNa, &iflag) == 0) {
        cerr << "Fatal Error: " << argv[2] <<
            " is an empty file! Exit.\n\n";
        exit(-1);
    }
    // iflag is first set in dbasef1: if the residue has no structure,
    // iflag 
    // is set to 1, otherwise it is set to 3
    // here, explicitly don't use the structure information by setting all 
    // iflags
    // to 1.
    if (intStructureFlag == 0) {
        for (i = 0; i < dbase[0].T; i++)
            iflag[i] = 1;
    }
    // read GCA format

    cout << setw(7) << NNa << " sequences read from " << argv[2] << "\n" <<
        flush;
    cout << " Length = " << setw(7) << dbase[0].T << "\n" << flush;
    // for (i = 0; i < dbase[0].T; i++) {
        // fix iflag = 1
        // iflag[i] = 1;
//        cout << dbase[0].resSeq[i] << " " << iflag[i] << endl << flush;
    // }

    // dcompress ( &mod ) ; 
    intrans_get(&intrans, mod);
    outtrans_get(&outtrans, mod);

    // ofstream out( "gamma.dat" ) ;
    ofstream        out(argv[3]);
    // cout << " Writing to " << argv[3] << endl << flush;

    // for ( na=0 ; na<Na ; na++ ){

    T = dbase[na].T;

    if (dbase[na].domainID == ' ')
        dbase[na].domainID = '_';
    if (!isalnum(dbase[na].domainID))
        dbase[na].domainID = '_';
    if (dbase[na].chainID == ' ')
        dbase[na].chainID = '_';
    if (!isalnum(dbase[na].chainID))
        dbase[na].chainID = '_';

    // Yu Shao Jan 29, 2004
    // The correct prob_flag value:
    // = 1 when don't use structure
    // = 3 when use structure
    if (intStructureFlag) {
        out << "HMM: " << argv[1] << ", prob_flag= 3 num_nodes= " << N <<
            endl;
        out << "REMARK: Structure was used, where present." <<
            endl;
    } else {
        out << "HMM: " << argv[1] << ", prob_flag= 1 num_nodes= " << N <<
            endl;
        out << "REMARK: Structure was NOT used, ignored if present." <<
            endl;
    }
    out << "GCA:  " << dbase[na].code << " " << dbase[na].chainID <<
        " " << dbase[na].domainID << " " << T << " " <<
        "Filename= " << argv[2] << endl << flush;
    if (argc == 7)
        out << "CATH: " << argv[5] << " " << argv[6] << endl;

    ct = new double[T];
    gamma = new double *[T];
    alpha = new double *[T];
    beta = new double *[T];
    for (t = 0; t < T; t++) {
        gamma[t] = new double[N];
        alpha[t] = new double[N];
        beta[t] = new double[N];
    }
    for (t = 0; t < T; t++) {
        ct[t] = 0.;
        for (i = 0; i < N; i++) {
            alpha[t][i] = 0.;
            beta[t][i] = 0.;
        }
    }

    // forward algorithm
    getalpha_brdc_prod(alpha, dbase[na], mod, intrans, ct, iflag);
    // backward algorithm
    getbeta_brdc_prod(beta, dbase[na], mod, outtrans, ct, iflag);

    // Calculate gamma matrix 

    for (t = 0; t < T; t++) {
        for (i = 0; i < N; i++) {
            gamma[t][i] = alpha[t][i] * beta[t][i] / ct[t];
        }
    }

    // debug: sanity test of gamma
    for (t = 0; t < T; t++) {
        double          norm = 0;
        for (i = 1; i < N; i++) {
            norm += gamma[t][i];
        }
        for (i = 1; i < N; i++) {
            gamma[t][i] /= norm;
        }
        // cout << "Gamma sum at residue " << t + 1 << " is: " << junk <<
        // endl;
    }

    q_ptr_r = new int[T];
    cfr = new float[T];

    qlambda_ramaland(dbase[na], mod, q_ptr_r, cfr, gamma);


    // Yu Shao 20-FEB-2002
    // calculate the secondary structure distribution for
    // each residue position
    // 1. allocate memory
    ss_prob = new double *[T];
    ss_max_prob = new char[T];
    for (t = 0; t < T; t++) {
        ss_prob[t] = new double[MDSSP];
    }
    for (t = 0; t < T; t++) {
        for (i = 0; i < MDSSP; i++) {
            ss_prob[t][i] = 0.;
        }
        ss_max_prob[t] = '\0';
    }
    // 2. calculate ss probability
    for (t = 0; t < T; t++) {
        for (intSS = 0; intSS < MDSSP; intSS++) {
            for (intQ = 1; intQ < N; intQ++) {
                ss_prob[t][intSS] += mod.d[intQ][intSS] * gamma[t][intQ];
            }
        }
    }
    // 3. indentifiy the most probable ss for each position
    for (t = 0; t < T; t++) {
        j = 0;
        for (i = 1; i < MDSSP; i++) {
            if (ss_prob[t][i] > ss_prob[t][j])
                j = i;
        }
        ss_max_prob[t] = mod.ss_map[j];
    }


    if (verbose) {
      for (t = 0; t < T; t++) {
          // Observed structure
          cout << setw(5) << t + 1 << " ";
          cout << setw(5) << *(dbase[na].resSeq + t) << " ";
          cout << *(dbase[na].O_ptr + t) << " " << *(dbase[na].O_rama + t)
              << *(dbase[na].O_ss + t) << *(dbase[na].O_context + t) << " ";
          cout << setw(8) << setprecision(2) << setiosflags(ios::fixed)
              << *(dbase[na].O_phi + t);
          cout << setw(8) << setprecision(2) << setiosflags(ios::fixed)
              << *(dbase[na].O_psi + t);
  
          // Predicted structure
          cout << setw(8) << setprecision(2) << setiosflags(ios::fixed) << ramacen[*(q_ptr_r + t)][0];        // angles
          cout << setw(8) << setprecision(2) << setiosflags(ios::fixed)
              << ramacen[*(q_ptr_r + t)][1] << " ";
  
          cout << rama_string[*(q_ptr_r + t)] << " ";        // rama region r
  
          cout << setw(8) << setprecision(5) << setiosflags(ios::fixed);
          cout << cfr[t];
          x = fabs(*(dbase[na].O_phi + t) - ramacen[*(q_ptr_r + t)][0]);
          if (x > 180.)
              x = 360. - x;
          z = fabs(*(dbase[na].O_psi + t) - ramacen[*(q_ptr_r + t)][1]);
          if (z > 180.)
              z = 360. - z;
          if (x < MDACUT && z < MDACUT) {
              cout << " true";
          } else {
              cout << " false";
          }
          cout << endl << flush;
          cout.unsetf(ios::fixed);
  
      }
    }

    if (!isalnum(dbase[na].O_context[t]))
        dbase[na].O_context[t] = '_';
    for (t = 0; t < T; t++) {
        out << "RESIDUE: ";
        out << t + 1 << " " << dbase[na].resSeq[t] << " ";
        out << dbase[na].O_ptr[t] << " ";
        out << setw(8) << setprecision(1) << setiosflags(ios::fixed);
        out << dbase[na].O_phi[t] << " ";
        out << setw(8) << setprecision(1) << setiosflags(ios::fixed);
        out << dbase[na].O_psi[t] << " ";
        out << setw(8) << setprecision(1) << setiosflags(ios::fixed);
        out << dbase[na].O_omg[t] << " ";
        out << setw(8) << setprecision(2) << setiosflags(ios::fixed);
        out << dbase[na].ca[t][0] << " ";
        out << setw(8) << setprecision(2) << setiosflags(ios::fixed);
        out << dbase[na].ca[t][1] << " ";
        out << setw(8) << setprecision(2) << setiosflags(ios::fixed);
        out << dbase[na].ca[t][2] << " ";
        out << dbase[na].O_ss[t] << " " << dbase[na].O_rama[t] << " ";
        out << dbase[na].O_context[t] << endl << flush;

        // Yu Shao 20-FEB-2003
        // Secondary structure
        out << "SS: ";
        out << ss_max_prob[t] << " ";
        // dSum = 0.;
        for (i = 0; i < MDSSP; i++) {
            out << setw(8) << setprecision(6) << setiosflags(ios::fixed);
            out << ss_prob[t][i] << " ";
            // dSum += ss_prob[t][i];
        }
        // NOTE: dSum should be zero, but it is +/-0.001
        // cout << dSum << endl;
        out << endl;

        // profile
        out << "PROFILE: ";
        for (i = 0; i < 20; i++) {
            out << setw(8) << setprecision(6) << setiosflags(ios::fixed) <<
                dbase[na].O_profile[t][i] << " ";
        }
        out << endl << flush;

        // optional: emission profile from gamma
        if (GPROF) {
            out << "GPROFILE: ";
            for (i = 0; i < 20; i++) {
                x = 0.;
                g = 0.;
                for (j = 0; j < N; j++) {
                    g += gamma[t][j];
                    x += gamma[t][j] * mod.b[j][i];
                }
                x /= g;
                out << setw(8) << setprecision(6) << setiosflags(ios::
                                                                 fixed) <<
                    x << " ";
            }
            out << endl << flush;
        }
        // gamma values
        out << "GAMMA: ";
        for (i = 0; i < N; i++) {
            out << setw(8) << setprecision(6) << setiosflags(ios::fixed) <<
                gamma[t][i] << " ";
        }
        out << endl << flush;

        // now output distance matrix
        out << "DISTANCE: ";
        y[0] = dbase[na].ca[t][0];
        y[1] = dbase[na].ca[t][1];
        y[2] = dbase[na].ca[t][2];
        for (i = 0; i < T; i++) {
            if (iflag[i] == 3) {
                x = 0;
                z = y[0] - dbase[na].ca[i][0];
                x += z * z;
                z = y[1] - dbase[na].ca[i][1];
                x += z * z;
                z = y[2] - dbase[na].ca[i][2];
                x += z * z;
                x = sqrt(x);
                // if (x < CONTACT) {out << "1 "; } else {out << "0 ";}
                // out << setw(6) << setprecision(1) <<
                // setiosflags(ios::fixed) <<
                // x << " " ;
            } else {
                x = 0;
            }
            ii = int        (x);
            out << ii << " ";
        }
        // end of distance matrix output
        out << endl << flush;
    }

    for (t = 0; t < T; t++) {
        delete          alpha[t];
        delete          beta[t];
        delete[]ss_prob[t];
    }
    delete          alpha;
    delete          beta;
    delete[]ss_prob;
    delete[]ss_max_prob;
}
