//*=========================================================================
//** Copyright 2003 Chris Bystroff
//**
//** Permission is hereby granted to use, execute, copy, distribute, modify,
//** and distribute in modified form this software for non-profit purposes,
//** provided this notice is retained in any and all copies.
//**
//** For for-profit usage, please contact bystrc@rpi.edu
//*=========================================================================

// Warning: DSSP symbols HEGST_ 
// Three state reduction :
// OLD convention H->H E->E {GST_}->L
// New convention (consistent with Rost etc. ) {H,G}->H E->E {ST_}->L
// Maps below that are consistent with OLD convention:
// threestate (modified 3/19/99)
// dcompress (modified 3/19/99)
// inv_dcompress (created 9/17/98)

#include <stdlib.h>
#include <math.h>
//#include "hmm.h"
//#include "ss_maps.h"

/***************************************************************************
 inv_dcompress.c - inversion of dcompress dssp symbol order 
                   all L are assigned to T
****************************************************************************/

void inv_dcompress ( struct model *modl ) 
{
  int i ; 
  
  modl->dground[0] = modl->dground[0] ;
  modl->dground[1] = modl->dground[1] ;
  modl->dground[4] = modl->dground[2] ; 
  modl->dground[2] = 0. ;
  modl->dground[3] = 0. ; 
  modl->dground[5] = 0. ;

  for ( i=0 ; i<modl->N ; i++ ){ 
    modl->d[i][0] = modl->d[i][0] ;
    modl->d[i][1] = modl->d[i][1] ;
    modl->d[i][4] = modl->d[i][2] ;
    modl->d[i][2] = 0. ; 
    modl->d[i][3] = 0. ;
    modl->d[i][5] = 0. ;    
  }

}

/****************************************************************************
  dcompress.c -  convert mod.d distribution to HEL 
****************************************************************************/
// modified 19-MAR-99 to map G to loop.

void dcompress ( struct model *modl ) 
{
  int i ; 

  // This is the Rost convention:
  // modl->dground[0] = modl->dground[0] + modl->dground[2] ;
  // modl->dground[1] = modl->dground[1] ;
  // modl->dground[2] = modl->dground[3] + modl->dground[4] + modl->dground[5] ; 
  modl->dground[0] = modl->dground[0] ;
  modl->dground[1] = modl->dground[1] ;
  modl->dground[2] = modl->dground[2] + modl->dground[3] 
                   + modl->dground[4] + modl->dground[5] ; 
  modl->dground[3] = 0. ; 
  modl->dground[4] = 0. ;
  modl->dground[5] = 0. ;

  for ( i=0 ; i<modl->N ; i++ ){ 
    modl->d[i][0] = modl->d[i][0] ;
    modl->d[i][1] = modl->d[i][1] ;
    modl->d[i][2] = modl->d[i][2] + modl->d[i][3] + 
                    modl->d[i][4] + modl->d[i][5] ; 
    modl->d[i][3] = 0. ; 
    modl->d[i][4] = 0. ;
    modl->d[i][5] = 0. ;    
  }

}

/***********************************************************************/

void d6to3 ( float *a, float *b ){

  b[0] = a[0] ;  // H 
  b[1] = a[1] ;  // E 
  b[2] = a[2] + a[3] + a[4] + a[5] ;  // G,S,T,_ -> L

}

/************************************************************************/

int irama ( char rama_symbol ){

  int i;
#ifdef USE_NEW_RAMA_MAP
  char rama_string[MRAMA+1] = "HGBEdbeLlxc" ;
#else
  char rama_string[MRAMA+1] = "HELX" ;
#endif

  for ( i=0 ; i<MRAMA ; i++ ){
    if ( rama_symbol == rama_string[i] ) return i;
  }
  return -1 ;
  
}

/***************************************************************************/

// 7-27-98 Converted to Rost etc convention
// 19-mar-99 'G' is now mapped to L-state

int threestate ( char dsspsymbol ) {

  if ( dsspsymbol == 'H' )
    return 0 ; 
  if ( dsspsymbol == 'G' || dsspsymbol == 'I' )
    return 2 ; 
  if ( dsspsymbol == 'E' ) 
    return 1 ; 
  if ( dsspsymbol == 'S' || dsspsymbol == 'T' || \
       dsspsymbol == 'B' || dsspsymbol == '_' )
    return 2 ; 
 
  return -1 ; 
}

/***************************************************************************/
int sixstate ( char dsspsymbol ) {

  if ( dsspsymbol == 'H' )
    return 0 ; 
  if ( dsspsymbol == 'E' ) 
    return 1 ; 
  if ( dsspsymbol == 'G' )
    return 2 ; 
  if ( dsspsymbol == 'S' )
    return 3 ; 
  if ( dsspsymbol == 'T' )
    return 4 ; 
  if ( dsspsymbol == 'I' || dsspsymbol == 'B' || \
       dsspsymbol == '_' )
    return 5 ; 
 
  return -1 ; 
}

/***************************************************************************/

int icontext ( char context_symbol ){

  int i; 
  char context_string[MCTXT+1] = "amnhdstul_" ;

  for ( i=0 ; i<MCTXT ; i++ ){
    if ( context_symbol == context_string[i] ) return i ;
  }
  return -1 ;

}


/***************************************************************************/

#ifdef USE_NEW_RAMA_MAP
char  Ramatype(float ph, float ps, float om)
{
  // char ramalett[12] = "LbdlBHEFxec";
  char rama_string[MRAMA+1] = "HGBEdbeLlxc" ;
  float d,dmin,ds,df;
  int i, icen;
  extern float ramacen[11][2];

  if (ph == 0.) return('?') ; // first residue
  if (ps == 0.) return('?') ; // last residue
  if (om < 90. && om > -90.) return 'c';

  dmin = 999.;
  for (i=0;i<MRAMA-1;i++) {
    df = fabs(ph - ramacen[i][0]); if (df>180.) df = 360.-df;
    ds = fabs(ps - ramacen[i][1]); if (ds>180.) ds = 360.-ds;
    d = sqrt(ds*ds + df*df);
    if (d < dmin) { icen=i; dmin=d; }
  }
  return rama_string[icen] ;
}
#else
char  Ramatype(float ph, float ps, float om)
{

  if (ph == 0.) return('?') ; // first residue
  if (ps == 0.) return('?') ; // last residue
  if (om < 90. && om > -90.) return('?') ; // cis peptide, set to unknown

    if (ph > 0. ) {
      if (ps > -90. && ps < 90.) {
        return('L') ; // left-handed helix
      } else {
        if (ph > 30. && ph < 150.) {
          return('X') ; //  rare
        } else {
          return('E') ; // fringe extended
        }
      }
    } else {
      if (ps < 40. && ps > -120.) {
        return('H') ; // helix
      } else {
        return('E') ; // extended
      }
    }
}
#endif

/**********************************************************************/

int iramachandran ( float phi, float psi ){
  if (phi == 0.) return -1;
  if (psi == 0.) return -1;
  if (phi > 0.) {
    if (psi > -90. && psi < 90.) {
      return 2 ;
    } else {
      if (phi > 30. && phi < 150.) {
        return 3 ;
      } else {
        return 1 ; 
      }
    }
  } else {
    if (psi < 40. && psi > -120.) {
      return 0 ;
    } else {
      return 1 ;
    }
  }
}
      
//  if ( phi<-2 && psi <30 && psi>-120 && psi!=0 )       // Helix 
//    return 0 ;                                
//  else if ( phi>.2 && phi<150 && psi>-120 && psi<120 )   // Loop
//    return 1 ; 
//  else if ( (phi>-.2 && phi <.2) || (psi==0 && phi<-2) )  // "Junk" 
//    return 3 ;
//  else         // Extended
//    return 2 ;
//}

/************************************************************************/

char ramachandran ( float phi, float psi ){
  if (phi == 0.) return '?';
  if (psi == 0.) return '?';
  if (phi > 0.) {
    if (psi > -90. && psi < 90.) {
      return 'L' ;
    } else {
      if (phi > 30. && phi < 150.) {
        return 'X' ;
      } else {
        return 'E' ; 
      }
    }
  } else {
    if (psi < 40. && psi > -120.) {
      return 'H' ;
    } else {
      return 'E' ;
    }
  }
}
// char ramachandran ( float phi, float psi ){
//   if ( phi<-2 && psi <30 && psi>-120 && psi!=0 )       // Helix 
//     return 'H';                                
//   else if ( phi>.2 && phi<150 && psi>-120 && psi<120 )   // Loop
//     return 'L'; 
//   else if ( (phi>-.2 && phi <.2) || (psi==0 && phi<-2) )  // "Junk" 
//     return 'J';
//   else         // Extended
//     return 'E' ;
// }

/*************************************************************************/

char tristate ( char ihssp ){
        if ( ihssp == 'H' || ihssp == 'G' ) 
                return 'H' ;
        else if ( ihssp == 'E' ) 
                return 'E' ;
        else 
                return 'L' ;
}






