// *=========================================================================
// ** Copyright 2003 Chris Bystroff
// **
// ** Permission is hereby granted to use, execute, copy, distribute,
// modify,
// ** and distribute in modified form this software for non-profit
// purposes,
// ** provided this notice is retained in any and all copies.
// **
// ** For for-profit usage, please contact bystrc@rpi.edu
// *=========================================================================
// 25-MAY-99 added ca (c_calpha)
// July 10, 1998 profile flag added, vt
// 28-MAY-98 cb
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
// #include "hmm.h"
// #include "vall_tools.h"
// #include "ss_maps.h"

#define MAXfamily 3000
#define Vmax 500000
// #define Vmax 140000
#define TOOSMALL 20
#define PROFFLAG 1
#define NEWDBASE 1
#define READCOORDS 1

// 
// low_filter - set observed frequences in profiles to 0
// if they are below a given threshold 
// 
#define THRESHOLD 0.1

void
low_filter(struct parent **dbase, int *NNa)
{
    int             na,
                    t,
                    T,
                    i,
                    ncount = 0,
	Tsum = 0,
	subcount;
    float           renorm;
    struct parent  *mydb;

    mydb = *dbase;

    cout << "Filtering out low count observations" << endl;
    for (na = 0; na < *NNa; na++) {
	T = mydb[na].T;
	Tsum += T;
	for (t = 0; t < T; t++) {
	    renorm = 1.;
	    subcount = 0;
	    for (i = 0; i < M; i++) {
		if (mydb[na].O_profile[t][i] < THRESHOLD &&
		    mydb[na].O_profile[t][i] > 0.) {
		    renorm -= mydb[na].O_profile[t][i];
		    mydb[na].O_profile[t][i] = 0.;
		    // cout << "removed: parent " << na << " position " << 
		    // t <<
		    // " aa " << i << endl;
		    subcount++;
		    ncount++;
		}
		if (subcount == 20) {
		    cout << "Error: all counts removed in parent " << na <<
			" position " << t << endl;
		    exit(8);
		}
	    }
	    // cout << "na " << na << " t " << t << " " << renorm << endl
	    // ;
	    for (i = 0; i < M; i++) {
		mydb[na].O_profile[t][i] /= renorm;
	    }
	}
    }
    cout << "Number of profile entries filtered out, per position: " <<
	float (ncount) / float (Tsum) << endl;

}

// 
// February 5, 1998
// 
// dbasef1.C - Reads database, returns protein sequences, dihedral angles, 
// 
// structure codes 
// ****************Reads only first protein (na = 1)*********
// 



int
dbasef1(char *vallfile, struct parent **dbase, int *NNa, int **iflag)
{

    int             i,
                    na,
                    count,
                    T,
                    ilen;
    short int       oldseq_list;
    char            code[MAXfamily][4];
    char            chainID[MAXfamily];
    int             seqlength[MAXfamily];
    char           *Otemp,
                   *sstemp;
    short int      *resseqtemp;
    float         **prof;
    float         **calpha;
    float          *phi,
                   *psi,
                   *omg;
    float          *gap_freq_temp,
                   *ins_freq_temp;
    char            domain[MAXfamily];
    char           *context;

    struct parent  *mydb;

    struct dbaserec rec;

    // allocate temporary storage
    resseqtemp = new short int[Vmax];
    Otemp = new char[Vmax];
    sstemp = new char[Vmax];
    phi = new float[Vmax];
    psi = new float[Vmax];
    omg = new float[Vmax];
    if (READCOORDS) {		// 25-MAY-99 cb
	calpha = new float *[Vmax];
	for (i = 0; i < Vmax; i++)
	    calpha[i] = new float[3];
    }
    if (PROFFLAG) {
	prof = new float *[Vmax];
	for (i = 0; i < Vmax; i++)
	    prof[i] = new float[M];
    }
    if (NEWDBASE) {
	gap_freq_temp = new float[Vmax];
	ins_freq_temp = new float[Vmax];
	context = new char[Vmax];
    }
    // 
    // open vallfile.. binary file of sequence + structure data.
    // 

    ifstream        in(vallfile, ios::in);
    if (in.bad()) {
	cerr << "Unable to open database file: " << vallfile << "\n";
	exit(8);
    }
    // 
    // Read whole file into temporary buffer 
    // 

    count = 0;
    na = 0;
    // compile error (Yu Shao Oct 30, 2003)
    // do { in.read( (unsigned char *) &rec, sizeof( rec ) ); } 
    do {
	in.read((char *) &rec, sizeof(rec));
    }
    while (rec.seq_resid == '-');
    while (in.gcount() == sizeof(rec)) {
      // start a new protein chain
      for (i=0;i<4;i++) code[na][i] = rec.code[i] ; 
      // code[na][4] = '\n';
      chainID[na] = rec.chainID;
      if (NEWDBASE) {
	domain[na] = rec.domainID;
      }
      ilen = 0;
      oldseq_list = -9999;
      // while ( abs(rec.resSeq) > oldseq_list ){
      while (in.gcount() == sizeof(rec)) {
	if (rec.seq_resid == 'Z')
	    rec.seq_resid = 'Q';
	if (rec.seq_resid == 'B')
	    rec.seq_resid = 'N';
	resseqtemp[count] = rec.resSeq;
	Otemp[count] = rec.seq_resid;
	phi[count] = rec.dih[0];
	psi[count] = rec.dih[1];
	omg[count] = rec.dih[2];
	sstemp[count] = rec.ss;
	if (READCOORDS) {
	    calpha[count][0] = rec.xyzca[0];
	    calpha[count][1] = rec.xyzca[1];
	    calpha[count][2] = rec.xyzca[2];
	}
	if (PROFFLAG) {
	    for (i = 0; i < M; i++) {
		prof[count][i] = rec.profile[i];
	    }
	}
	if (NEWDBASE) {
	    gap_freq_temp[count] = rec.gap_freq;
	    ins_freq_temp[count] = rec.ins_freq;
	    context[count] = rec.context;
	}
	++count;		// no. excluding deletions
	++ilen;			// length of chain 
	oldseq_list = abs(rec.resSeq);	// no. ungapped sequences
	// read in the next record
	// do { in.read( (unsigned char *) &rec, sizeof( rec ) ); } 
	do {
	    in.read((char *) &rec, sizeof(rec));
	}
	while (rec.seq_resid == '-');
	if (rec.seq_list == 1)
	    break;
      }
      seqlength[na] = ilen;
      cout << "  database protein " << na << " length=" << ilen << endl;
      ++na;
      break;  // one chain only
    } 
    *NNa = na ;
    in.close();

    // 
    // Allocate dbase and transfer everything.
    // 

    *dbase = new parent[*NNa];
    mydb = *dbase;
    cout << "Allocating space for " << *NNa << " chains.\n";
    cout << "Total number of residues= " << count << " \n";
    count = 0;
    na = 0;
    // for (na=0 ; na < *NNa; na++) {
      T = seqlength[na];
      *iflag = new int[T];
      mydb[na].T = T;
      for (i = 0; i < 4; i++)
	mydb[na].code[i] = code[na][i];
      mydb[na].chainID = chainID[na];
      if (NEWDBASE) {
	mydb[na].domainID = domain[na];
      }
      // allocate sub-arrays
      mydb[na].O_ptr = new char[T];
      mydb[na].O_phi = new float[T];
      mydb[na].O_psi = new float[T];
      mydb[na].O_omg = new float[T];
      mydb[na].O_ss = new char[T];
      mydb[na].O_rama = new char[T];
      mydb[na].resSeq = new short int[T];
      if (READCOORDS) {		// 25-MAY-99 cb
	mydb[na].ca = new float *[T];
	for (i = 0; i < T; i++)
	    mydb[na].ca[i] = new float[3];
      }
      if (PROFFLAG) {
	mydb[na].O_profile = new float *[T];
	for (i = 0; i < T; i++)
	    mydb[na].O_profile[i] = new float[M];
      }
      if (NEWDBASE) {
	mydb[na].O_gap = new float[T];
	mydb[na].O_ins = new float[T];
	mydb[na].O_context = new char[T];
      }
      // transfer from temporary memory
      substring(Otemp + count, mydb[na].O_ptr, T);
      substring2(phi + count, mydb[na].O_phi, T);
      substring2(psi + count, mydb[na].O_psi, T);
      substring2(omg + count, mydb[na].O_omg, T);
      substring(sstemp + count, mydb[na].O_ss, T);
      // substring3 ( resseqtemp+count,mydb[na].resSeq , T ) ; 
      for (i = 0; i < T; i++) {
	// if (*(resseqtemp+count+i) > 0 ) 
	if ((*(phi + count + i) == 0.000) || (*(psi + count + i) == 0.000)) {
	    (*iflag)[i] = 1;
	} else {
	    (*iflag)[i] = 3;
	}
	*(mydb[na].resSeq + i) = abs(*(resseqtemp + count + i));
      }
      // convert angles to Rama character: HELX
      for (i = 0; i < T; i++)
	*(mydb[na].O_rama + i) =
	    Ramatype(mydb[na].O_phi[i], mydb[na].O_psi[i],
		     mydb[na].O_omg[i]);
      // save coordinates
      if (READCOORDS) {
	for (i = 0; i < T; i++) {
	    substring2(calpha[i + count], mydb[na].ca[i], 3);
	}
      }
      // save profile data
      if (PROFFLAG) {
	for (i = 0; i < T; i++) {
	    substring2(prof[i + count], mydb[na].O_profile[i], 20);
	}
      }

      if (NEWDBASE) {
	substring2(gap_freq_temp + count, mydb[na].O_gap, T);
	substring2(ins_freq_temp + count, mydb[na].O_ins, T);
	substring(context + count, mydb[na].O_context, T);
      }
      count += seqlength[na];
    // }
    cout << "Total number of residues stored in mem= " << count << " \n";
    cout << "Last chain " << setw(6) << na << " : ";
    for (i = 0; i < 4; i++)
	// don't use "na - 1" here -- na is 0. Yu Shao Feb 9 2003
	// cout << mydb[na].code[i];
	cout << mydb[na - 1].code[i];
    cout << "\n";

    // Clean up temporary storage

    delete[]Otemp;
    Otemp = NULL;
    delete[]sstemp;
    sstemp = NULL;
    delete[]phi;
    phi = NULL;
    delete[]psi;
    psi = NULL;
    delete[]omg;
    omg = NULL;
    delete[]resseqtemp;
    resseqtemp = NULL;
    if (READCOORDS) {
	for (i = 0; i < Vmax; i++)
	    delete[]calpha[i];
	delete[]calpha;
	calpha = NULL;
    }

    if (PROFFLAG) {
	for (i = 0; i < Vmax; i++)
	    delete[]prof[i];
	delete[]prof;
	prof = NULL;
    }
    // return 0 if the drct file is empty
    // -- by Yu Shao (Oct 30 2003)
    return ilen;

}

// 
// February 5, 1998
// 
// dbasef.C - Reads database, returns protein sequences, dihedral angles, 
// 
// structure codes 
// ****************Reads multiple proteins *******************
// 



int
dbasef(char *vallfile, struct parent **dbase, int *NNa)
{

    int             i,
                    na,
                    count,
                    T,
                    ilen;
    short int       oldseq_list;
    char            code[MAXfamily][4];
    char            chainID[MAXfamily];
    int             seqlength[MAXfamily];
    char           *Otemp,
                   *sstemp;
    short int      *resseqtemp;
    float         **prof;
    float         **calpha;
    float          *phi,
                   *psi,
                   *omg;
    float          *gap_freq_temp,
                   *ins_freq_temp;
    char            domain[MAXfamily];
    char           *context;

    struct parent  *mydb;

    struct dbaserec rec;

    // allocate temporary storage
    resseqtemp = new short int[Vmax];
    Otemp = new char[Vmax];
    sstemp = new char[Vmax];
    phi = new float[Vmax];
    psi = new float[Vmax];
    omg = new float[Vmax];
    if (READCOORDS) {		// 25-MAY-99 cb
	calpha = new float *[Vmax];
	for (i = 0; i < Vmax; i++)
	    calpha[i] = new float[3];
    }
    if (PROFFLAG) {
	prof = new float *[Vmax];
	for (i = 0; i < Vmax; i++)
	    prof[i] = new float[M];
    }
    if (NEWDBASE) {
	gap_freq_temp = new float[Vmax];
	ins_freq_temp = new float[Vmax];
	context = new char[Vmax];
    }
    // 
    // open vallfile.. binary file of sequence + structure data.
    // 

    ifstream        in(vallfile, ios::in);
    if (in.bad()) {
	cerr << "Unable to open database file: " << vallfile << "\n";
	exit(8);
    }
    // 
    // Read whole file into temporary buffer 
    // 

    count = 0;
    na = 0;
    // compile error (Yu Shao Oct 30, 2003)
    // do { in.read( (unsigned char *) &rec, sizeof( rec ) ); } 
    do {
	in.read((char *) &rec, sizeof(rec));
    }
    while (rec.seq_resid == '-');
    while (in.gcount() == sizeof(rec)) {
      // start a new protein chain
      for (i=0;i<4;i++) code[na][i] = rec.code[i] ; 
      // code[na][4] = '\n';
      chainID[na] = rec.chainID;
      if (NEWDBASE) {
	domain[na] = rec.domainID;
      }
      ilen = 0;
      oldseq_list = -9999;
      // while ( abs(rec.resSeq) > oldseq_list ){
      while (in.gcount() == sizeof(rec)) {
	if (rec.seq_resid == 'Z')
	    rec.seq_resid = 'Q';
	if (rec.seq_resid == 'B')
	    rec.seq_resid = 'N';
	resseqtemp[count] = rec.resSeq;
	Otemp[count] = rec.seq_resid;
	phi[count] = rec.dih[0];
	psi[count] = rec.dih[1];
	omg[count] = rec.dih[2];
	sstemp[count] = rec.ss;
	if (READCOORDS) {
	    calpha[count][0] = rec.xyzca[0];
	    calpha[count][1] = rec.xyzca[1];
	    calpha[count][2] = rec.xyzca[2];
	}
	if (PROFFLAG) {
	    for (i = 0; i < M; i++) {
		prof[count][i] = rec.profile[i];
	    }
	}
	if (NEWDBASE) {
	    gap_freq_temp[count] = rec.gap_freq;
	    ins_freq_temp[count] = rec.ins_freq;
	    context[count] = rec.context;
	}
	++count;		// no. excluding deletions
	++ilen;			// length of chain 
	oldseq_list = abs(rec.resSeq);	// no. ungapped sequences
	// read in the next record
	// do { in.read( (unsigned char *) &rec, sizeof( rec ) ); } 
	do {
	    in.read((char *) &rec, sizeof(rec));
	}
	while (rec.seq_resid == '-');
	if (rec.seq_list == 1)
	    break;
      }
      seqlength[na] = ilen;
      cout << "  database protein " << na << " length=" << ilen << endl;
      ++na;
    } 
    *NNa = na ;
    in.close();

    // 
    // Allocate dbase and transfer everything.
    // 

    *dbase = new parent[*NNa];
    mydb = *dbase;
    cout << "Allocating space for " << *NNa << " chains.\n";
    cout << "Total number of residues= " << count << " \n";
    count = 0;
    for (na=0 ; na < *NNa; na++) {
      T = seqlength[na];
      // *iflag = new int[T];
      mydb[na].T = T;
      for (i = 0; i < 4; i++)
	mydb[na].code[i] = code[na][i];
      mydb[na].chainID = chainID[na];
      if (NEWDBASE) {
	mydb[na].domainID = domain[na];
      }
      // allocate sub-arrays
      mydb[na].O_ptr = new char[T];
      mydb[na].O_phi = new float[T];
      mydb[na].O_psi = new float[T];
      mydb[na].O_omg = new float[T];
      mydb[na].O_ss = new char[T];
      mydb[na].O_rama = new char[T];
      mydb[na].resSeq = new short int[T];
      if (READCOORDS) {		// 25-MAY-99 cb
	mydb[na].ca = new float *[T];
	for (i = 0; i < T; i++)
	    mydb[na].ca[i] = new float[3];
      }
      if (PROFFLAG) {
	mydb[na].O_profile = new float *[T];
	for (i = 0; i < T; i++)
	    mydb[na].O_profile[i] = new float[M];
      }
      if (NEWDBASE) {
	mydb[na].O_gap = new float[T];
	mydb[na].O_ins = new float[T];
	mydb[na].O_context = new char[T];
      }
      // transfer from temporary memory
      substring(Otemp + count, mydb[na].O_ptr, T);
      substring2(phi + count, mydb[na].O_phi, T);
      substring2(psi + count, mydb[na].O_psi, T);
      substring2(omg + count, mydb[na].O_omg, T);
      substring(sstemp + count, mydb[na].O_ss, T);
      // substring3 ( resseqtemp+count,mydb[na].resSeq , T ) ; 
      for (i = 0; i < T; i++) {
	// if (*(resseqtemp+count+i) > 0 ) 
	// if ((*(phi + count + i) == 0.000) || (*(psi + count + i) == 0.000)) {
	//     (*iflag)[i] = 1;
	// } else {
	//     (*iflag)[i] = 3;
	// }
	*(mydb[na].resSeq + i) = abs(*(resseqtemp + count + i));
      }
      // convert angles to Rama character: HELX
      for (i = 0; i < T; i++)
	*(mydb[na].O_rama + i) =
	    Ramatype(mydb[na].O_phi[i], mydb[na].O_psi[i],
		     mydb[na].O_omg[i]);
      // save coordinates
      if (READCOORDS) {
	for (i = 0; i < T; i++) {
	    substring2(calpha[i + count], mydb[na].ca[i], 3);
	}
      }
      // save profile data
      if (PROFFLAG) {
	for (i = 0; i < T; i++) {
	    substring2(prof[i + count], mydb[na].O_profile[i], 20);
	}
      }

      if (NEWDBASE) {
	substring2(gap_freq_temp + count, mydb[na].O_gap, T);
	substring2(ins_freq_temp + count, mydb[na].O_ins, T);
	substring(context + count, mydb[na].O_context, T);
      }
      count += seqlength[na];
    }
    cout << "Total number of residues stored in mem= " << count << " \n";
    cout << "Last chain " << setw(6) << na << " : ";
    for (i = 0; i < 4; i++)
	// don't use "na - 1" here -- na is 0. Yu Shao Feb 9 2003
	// cout << mydb[na].code[i];
	cout << mydb[na - 1].code[i];
    cout << "\n";

    // Clean up temporary storage

    delete[]Otemp;
    Otemp = NULL;
    delete[]sstemp;
    sstemp = NULL;
    delete[]phi;
    phi = NULL;
    delete[]psi;
    psi = NULL;
    delete[]omg;
    omg = NULL;
    delete[]resseqtemp;
    resseqtemp = NULL;
    if (READCOORDS) {
	for (i = 0; i < Vmax; i++)
	    delete[]calpha[i];
	delete[]calpha;
	calpha = NULL;
    }

    if (PROFFLAG) {
	for (i = 0; i < Vmax; i++)
	    delete[]prof[i];
	delete[]prof;
	prof = NULL;
    }
    // return 0 if the drct file is empty
    // -- by Yu Shao (Oct 30 2003)
    return ilen;

}

// 
// July 27, 1998
// 
// dbasef_ascii.C - Reads database, returns protein sequences, dihedral
// angles, 
// structure codes 
// vall file in ascii format 
// 

#define MAXfamily2 1500
#define Vmax2 300000

/***************************************************************************
   get_vall_record.c : Get single ascii format vall record 
***************************************************************************/

void
get_vall_record(struct dbaserec_ascii *record, FILE * fp)
{

    int             i,
                    nwords;
    char            line[200];

    fgets(line, sizeof(line), fp);
    for (i = 0; i < (M + 1); i++) {
	sscanf(line + 1 + 6 * i, " %f", &(record->profile[i]));
    }

    fgets(line, sizeof(line), fp);
    nwords = wc_string(line);

    if (nwords == 14)
	sscanf(line, "%c%i%i%i %8f%8f%8f%8f%8f%8f%8f%s %c%i",
	       &(record->ss), &(record->seq_down), &(record->seq_list),
	       &(record->resSeq), &(record->xyzca[0]), &(record->xyzca[1]),
	       &(record->xyzca[2]), &(record->dih[0]), &(record->dih[1]),
	       &(record->dih[2]), &(record->chi), (record->code),
	       &(record->seq_resid), &(record->nalign));
    else if (nwords == 12) {
	sscanf(line, "%i%i%i %8f%8f%8f%8f%8f%8f%8f%s %i",
	       &(record->seq_down), &(record->seq_list), &(record->resSeq),
	       &(record->xyzca[0]), &(record->xyzca[1]),
	       &(record->xyzca[2]), &(record->dih[0]), &(record->dih[1]),
	       &(record->dih[2]), &(record->chi), (record->code),
	       &(record->nalign));
	record->ss = ' ';
	record->seq_resid = ' ';
    }

    record->chainID = ' ';


}

/*************************************************************************
       Substring of string  of length N 
**********************************************************************/

void
substring(char *string, char *substr_ptr, short int N)
{
    int             i;

    for (i = 0; i < N; i++) {
	if (isalnum(*(string + i)))
	    *(substr_ptr + i) = *(string + i);
	else
	    *(substr_ptr + i) = '.';
    }
}

void
substring2(float *string, float *substr_ptr, short int N)
{

    int             i;

    for (i = 0; i < N; i++)
	*(substr_ptr + i) = *(string + i);

}
void
substring3(short int *string, short int *substr_ptr, short int N)
{

    int             i;

    for (i = 0; i < N; i++)
	*(substr_ptr + i) = *(string + i);

}

/***********************************************************************
   wc_string.c - count words in string 
   ( Kernighan&Ritchie, The c Programming Language, Section 1.5.4 ) 
*************************************************************************/


#define IN  1			/* inside a word */
#define OUT 0			/* outside a word */

int
wc_string(char *ptr)
{

    int             i,
                    c,
                    state,
                    nw;

    state = OUT;
    nw = 0;

    for (i = 0; i < int (strlen(ptr)); i++) {
	c = *(ptr + i);
	if (c == ' ' || c == '\t')
	    state = OUT;
	else if (state == OUT) {
	    state = IN;
	    ++nw;
	}
    }

    return nw;

}

/*************************************************************************/
float dist(float * x, float * y) {
  float d;
  int i;
  for (i=0, d=0.;i<3;++i) {
    d += (x[i] - y[i])*(x[i] - y[i]);
  }
  return sqrt(d);
}

/*************************************************************************/
float dotprod(float * v1, float * v2) {
  int i;
  float x;
  x = 0.;
  for (i=0; i<3; ++i) {
    x += v1[i]*v2[i];
  }
  return x;
}

/*************************************************************************/
void cross(float * v1,float * v2,float * v3) {
// !* v3 = v1 X v2
        v3[0] = v1[1]*v2[2] - v1[2]*v2[1];
        v3[1] = v1[2]*v2[0] - v1[0]*v2[2];
        v3[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

/*************************************************************************/
void subvec(float * v1,float * v2,float * v3) {
// !* v3 = v1 - v2 = vector from v2 to v1
  int i;
  for (i=0;i<3;++i) {
    v3[i] = v1[i] - v2[i];
  }
}

/*************************************************************************/
void unitvec(float * v,float * u) {
// !* u = v / |v|
  int i;
  float x;
  x = 0;
  for (i=0;i<3;++i) {
    x += v[i]*v[i];
  }
  x = sqrt(x);
  for (i=0;i<3;++i) {
    u[i] = v[i]/x;
  }
}

/*************************************************************************/
float atan2d(float y, float x) {
  // atan2d is the arctangent in degrees for
  // the vector x,y. atan() returns a value between
  // - pi/2 and pi/2. This is the arctan of y/x if x > 0.
  // But if x < 0, then pi should be added.
  //  if x == 0. then if y > 0, the angle is 90 deg, otherwise 270.
  float pi=3.1415927;
  float rad=180./pi;

  if (x == 0.) {
    if (y > 0) { return 90.; } else {return -90.;}
  } else {
    if (x > 0) { return rad*atan(y/x); } 
    else if (y > 0 ) { return rad*(pi + atan(y/x)); }
         else { return rad*(atan(y/x) - pi); }
  }
}

/*************************************************************************/
float measure(float * v1, float * v2, float * v3, float * v4) {
  float d, v21[3], v23[3], v34[3], vy[3], uy[3], vx[3], ux[3];
  int i;
  float cx,cy, cz;
  // cout << "Measuring: " ;
  for (i=0; i<3; ++i) {
   //  cout << " " << v1[i] << " " << v2[i] << " " << v3[i] << " " << v4[i] ;
    if (v1[i]==0.0) return 999.;
    if (v2[i]==0.0) return 999.;
    if (v3[i]==0.0) return 999.;
    if (v4[i]==0.0) return 999.;
  }
  // cout << endl;
  subvec(v3,v2,v23);
  // cout << "subvec: " << v23[0] << " " << v23[1] << " " << v23[2] << endl;
  subvec(v1,v2,v21);
  // cout << "subvec: " << v21[0] << " " << v21[1] << " " << v21[2] << endl;
  cross(v23,v21,vy);
  // cout << "cross: " << vy[0] << " " << vy[1] << " " << vy[2] << endl;
  unitvec(vy,uy);
  // cout << "unitvec: " << uy[0] << " " << uy[1] << " " << uy[2] << endl;
  cross(uy,v23,vx);
  // cout << "cross: " << vx[0] << " " << vx[1] << " " << vx[2] << endl;
  unitvec(vx,ux);
  // cout << "unitvec: " << ux[0] << " " << ux[1] << " " << ux[2] << endl;
  subvec(v4,v3,v34);
  // cout << "subvec: " << v34[0] << " " << v34[1] << " " << v34[2] << endl;
  cx = dotprod(v34,ux);
  // cout << "dotprod: " << cx << endl;
  cy = dotprod(v34,uy);
  // cout << "dotprod: " << cy << endl;
  cz = atan2d(cy,cx);
  // cout << "ATAN(" << cy << "/" << cx << ")=" << cz << endl;
  if (cz != cz) { return 999.; }
  return cz;
}

